/* Copyright (C) 2003-2006 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/
/*
#define DEBUG_GUESSER
*/
#include "dps_common.h"
#include "dps_hash.h"
#include "dps_guesser.h"
#include "dps_utils.h"
#include "dps_unicode.h"
#include "dps_log.h"
#include "dps_vars.h"
#include "dps_mutex.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <dirent.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>

/* This should be last include */
#ifdef DMALLOC
#include "dmalloc.h"
#endif

static int LangMapCmp(const DPS_LANGMAP *m1, const DPS_LANGMAP *m2) {
  register int r = strcasecmp(m1->lang, m2->lang);
  if (r != 0) return r;
  return strcasecmp(m1->charset, m2->charset);
}

static DPS_LANGMAP *FindLangMap(DPS_LANGMAPLIST *L, const char *lang, const char *charset, const char *filename, const int addnew) {
  DPS_LANGMAP *o = NULL;
  ssize_t i, l, r;
  register int c;
  char *lt, *language = dps_strtok_r(lang, ", ", &lt);

  if (L->nmaps) {
    while(language) {

      l = 0; r = (ssize_t)L->nmaps - 1;
      if (addnew) {
	while(l <= r) {
	  i = (l + r) / 2;
	  c = strcasecmp(L->Map[i].lang, language);
	  if (c == 0) {
	    c = strcasecmp(L->Map[i].charset, charset);
	  }
	  if (c == 0) return &L->Map[i];
	  if (c < 0) l = i + 1;
	  else r = i - 1;
	}

      }else {

	while(l <= r) {
	  i = (l + r) / 2;
	  c = strncasecmp(L->Map[i].lang, language, strlen(L->Map[i].lang));
	  if (c == 0) {
	    c = strncasecmp(L->Map[i].charset, charset, strlen(L->Map[i].charset));
	  }
	  if (c == 0) return &L->Map[i];
	  if (c < 0) l = i + 1;
	  else r = i - 1;
	}
      }
      language = dps_strtok_r(NULL, ", ", &lt);
    }
  }

  if (!addnew) return NULL;

  if(L->nmaps == 0){
    o = L->Map = (DPS_LANGMAP*)DpsMalloc(sizeof(DPS_LANGMAP));
    if (o == NULL) return NULL;
  }else{
    L->Map = (DPS_LANGMAP*)DpsRealloc(L->Map, (L->nmaps + 1) * sizeof(DPS_LANGMAP));
    if (L->Map == NULL) {
      L->nmaps = 0;
      return NULL;
    }
    o = &L->Map[L->nmaps];
  }
  if (o == NULL || L->Map == NULL) {
    fprintf(stderr, "Can't alloc/realloc for language map (%s, %s), nmaps: %d (%d)", lang, charset, 
		 L->nmaps + 1, (L->nmaps + 1) * sizeof(DPS_LANGMAP) );
    return NULL;
  }
  bzero((void*)o, sizeof(DPS_LANGMAP));
  for (i = 0; i <= DPS_LM_HASHMASK; i++) o->memb3[i].index = o->memb6[i].index = i;
  o->charset = (char*)DpsStrdup(charset);
  o->lang = (char*)DpsStrdup(lang);
  o->filename = (filename == NULL) ? NULL : (char*)DpsStrdup(filename);
  L->nmaps++;

/*  DpsPreSort(L->Map, L->nmaps, sizeof(DPS_LANGMAP), (qsort_cmp)LangMapCmp);*/
  DpsSort(L->Map, L->nmaps, sizeof(DPS_LANGMAP), (qsort_cmp)LangMapCmp);

  l = 0; r = (ssize_t)L->nmaps - 1;
  while(l <= r) {
    i = (l + r) / 2;
    c = strcasecmp(L->Map[i].lang, lang);
    if (c == 0) {
      c = strcasecmp(L->Map[i].charset, charset);
    }
    if (c == 0) return &L->Map[i];
    if (c < 0) l = i + 1;
    else r = i - 1;
  }
  return L->Map; /* this should never happen */
}


__C_LINK int __DPSCALL DpsLoadLangMapFile(DPS_LANGMAPLIST *L, const char * filename) {
     struct stat     sb;
     char *str, *cur_n = NULL, *data = NULL;
     char *Ccharset = NULL, *Clanguage = NULL;
     int Clen = DPS_LM_MAXGRAM1;
     DPS_LANGMAP *Cmap = NULL;
     DPS_CHARSET * cs;
     int             fd;
     char savebyte;

     if (stat(filename, &sb)) {
       fprintf(stderr, "Unable to stat LangMap file '%s': %s", filename, strerror(errno));
       return DPS_ERROR;
     }
     if ((fd = open(filename, O_RDONLY)) <= 0) {
       fprintf(stderr, "Unable to open LangMap file '%s': %s", filename, strerror(errno));
       return DPS_ERROR;
     }
     if ((data = (char*)DpsMalloc(sb.st_size + 1)) == NULL) {
       fprintf(stderr, "Unable to alloc %d bytes", sb.st_size);
       close(fd);
       return DPS_ERROR;
     }
     if (read(fd, data, sb.st_size) != (ssize_t)sb.st_size) {
       fprintf(stderr, "Unable to read LangMap file '%s': %s", filename, strerror(errno));
       DPS_FREE(data);
       close(fd);
       return DPS_ERROR;
     }
     data[sb.st_size] = '\0';
     str = data;
     cur_n = strchr(str, '\n');
     if (cur_n != NULL) {
       cur_n++;
       savebyte = *cur_n;
       *cur_n = '\0';
     }

     while(str != NULL) {
          if(str[0]=='#'||str[0]==' '||str[0]=='\t') goto loop_continue;

          if(!strncmp(str,"Charset:",8)){
               
               char * charset, * lasttok;
               DPS_FREE(Ccharset);
               if((charset = dps_strtok_r(str+8," \t\n\r",&lasttok))){
                       const char *canon = DpsCharsetCanonicalName(charset);
                    if (canon) {
                      Ccharset = (char*)DpsStrdup(canon);
                    } else {
                      fprintf(stderr, "Charset: %s in %s not supported\n", charset, filename);
                      return DPS_ERROR;
                    }
               }
          }else
          if(!strncmp(str,"Language:",9)){
               char * lang, *lasttok;
               DPS_FREE(Clanguage);
               if((lang = dps_strtok_r(str+9," \t\n\r",&lasttok))){
                    Clanguage = (char*)DpsStrdup(lang);
               }
          }else
          if(!strncmp(str,"Length:",7)){
               char * lang, *lasttok;
               if((lang = dps_strtok_r(str+9," \t\n\r",&lasttok))){
		 Clen = DPS_ATOI(lang);
               }
          }else{
               char *s;
               int count;
               
               if(!(s=strchr(str,'\t'))) goto loop_continue;
               if(Clanguage == NULL) {
                 fprintf(stderr, "No language definition in LangMapFile '%s'\n", filename);
                 return DPS_ERROR;
               }

               if(Ccharset == NULL) {
                 fprintf(stderr, "No charset definition in LangMapFile '%s'\n", filename);
                 return DPS_ERROR;
               }
               if(!(cs = DpsGetCharSet(Ccharset))) {
                 fprintf(stderr, "Unknown charset '%s' in LangMapFile '%s'\n", Ccharset, filename);
                 return DPS_ERROR;
               }
               if (Cmap == NULL) {
		 Cmap = FindLangMap(L, Clanguage, Ccharset, filename, 1);
		 DpsSort(Cmap->memb3, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
		 DpsSort(Cmap->memb6, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
	       }
               if (Cmap == NULL) return DPS_ERROR;
               *s='\0';

               if(((count = DPS_ATOI(s+1))==0)||(strlen(str)<1)/*||(strlen(str)>DPS_LM_MAXGRAM)*/)
                    goto loop_continue;

               if(*str){
                    int hindex;
		    sscanf(str, "%x", &hindex);
		    hindex &= DPS_LM_HASHMASK;
		    if (Clen == DPS_LM_MAXGRAM1) {
		      Cmap->memb3[hindex].count += count;
#ifdef DEBUG_GUESSER
		      strcpy(Cmap->memb3[hindex].str, str);
#endif
		    } else {
		      Cmap->memb6[hindex].count += count;
#ifdef DEBUG_GUESSER
		      strcpy(Cmap->memb6[hindex].str, str);
#endif
		    }
               }
          }
     loop_continue:
	  str = cur_n;
	  if (str != NULL) {
	    *str = savebyte;
	    cur_n = strchr(str, '\n');
	    if (cur_n != NULL) {
	      cur_n++;
	      savebyte = *cur_n;
	      *cur_n = '\0';
	    }
	  }
     }
     close(fd);
     DPS_FREE(data);
     DPS_FREE(Clanguage);
     DPS_FREE(Ccharset);
     
     if (Cmap != NULL) DpsPrepareLangMap(Cmap);
     return DPS_OK;
}

__C_LINK int __DPSCALL DpsLoadLangMapList(DPS_LANGMAPLIST *L, const char * mapdir) {

	DIR * dir;
	struct dirent * item;
	char fullname[1024]="";
	char name[1024]="";
	int res = DPS_OK;

	dir=opendir(mapdir);
	if(!dir)return 0;

	while((item=readdir(dir))){
		char * tail;
		strcpy(name,item->d_name);
		if((tail=strstr(name,".lm"))){
		        *tail='\0';
			sprintf(fullname,"%s/%s",mapdir,item->d_name);
			res = DpsLoadLangMapFile(L, fullname);
/*			if (res != DPS_OK) return res;*/
		}
	}
	closedir(dir);

	return DPS_OK;
}


void DpsLangMapListFree(DPS_LANGMAPLIST *List){
     size_t i;
     
     for(i=0;i<List->nmaps;i++){
          DPS_FREE(List->Map[i].charset);
          DPS_FREE(List->Map[i].lang);
          DPS_FREE(List->Map[i].filename);
     }
     DPS_FREE(List->Map);
     List->nmaps = 0;
}


void DpsLangMapListSave(DPS_LANGMAPLIST *List) {
     size_t i, j, minv;
     FILE *out;
     DPS_LANGMAP *Cmap;
     char name[128];
     time_t t=time(NULL);
     struct tm *tim = localtime(&t);
     char time_str[128];
     double ratio;
     
     for(i = 0; i < List->nmaps; i++) {
       Cmap = &List->Map[i];
       if (Cmap->needsave) {
         if (Cmap->filename == NULL) {
           dps_snprintf(name, 128, "%s.%s.lm", Cmap->lang, Cmap->charset);
           if ((out = fopen(name, "w")) == NULL) continue;
         } else {
           if ((out = fopen(Cmap->filename, "w")) == NULL) continue;
         }
	 strftime(time_str, sizeof(time_str), "%c %Z (UTC%z)", tim);
         fprintf(out, "#\n");
         fprintf(out, "# Autoupdated: %s, %s-%s\n", time_str, PACKAGE, VERSION);
         fprintf(out, "#\n\n");
         fprintf(out, "Language: %s\n", Cmap->lang);
         fprintf(out, "Charset:  %s\n", Cmap->charset);
         fprintf(out, "\n\n");

	 fprintf(out, "Length:   %d\n", DPS_LM_MAXGRAM1);
         DpsSort(Cmap->memb3, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpCount);
	 minv = (Cmap->memb3[DPS_LM_TOPCNT - 1].count > 8000) ? 8000 : Cmap->memb3[DPS_LM_TOPCNT - 1].count;
	 ratio = ((double) Cmap->memb3[DPS_LM_TOPCNT - 1].count) / minv;
	 if (ratio > 0.0) {
	   for(j = 0; j < DPS_LM_TOPCNT; j++) {
	     Cmap->memb3[j].count = (size_t) ((double)Cmap->memb3[j].count / ratio);
	   }
	 }
         for(j = 0; j < DPS_LM_TOPCNT; j++) {
          if(!Cmap->memb3[j].count) break;
          fprintf(out, "%03x\t%u\n", Cmap->memb3[j].index, Cmap->memb3[j].count);
         }

	 fprintf(out, "Length:   %d\n", DPS_LM_MAXGRAM2);
         DpsSort(Cmap->memb6, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpCount);
	 minv = (Cmap->memb6[DPS_LM_TOPCNT - 1].count > 8000) ? 8000 : Cmap->memb6[DPS_LM_TOPCNT - 1].count;
	 ratio = ((double) Cmap->memb6[DPS_LM_TOPCNT - 1].count) / minv;
	 if (ratio > 0.0) {
	   for(j = 0; j < DPS_LM_TOPCNT; j++) {
	     Cmap->memb6[j].count = (size_t) ((double)Cmap->memb6[j].count / ratio);
	   }
	 }
         for(j = 0; j < DPS_LM_TOPCNT; j++) {
          if(!Cmap->memb6[j].count) break;
          fprintf(out, "%03x\t%u\n", Cmap->memb6[j].index, Cmap->memb6[j].count);
         }

         fprintf(out, "#\n");
         fclose(out);
       }
     }
}



void DpsBuildLangMap(DPS_LANGMAP * map, const char * text, size_t textlen, size_t max_nbytes, int StrFlag) {
  unsigned char buf1[2 * DPS_LM_MAXGRAM1 + 1];
  unsigned char buf2[2 * DPS_LM_MAXGRAM2 + 1];
  const char * end = text + textlen;
  unsigned char code, prevb = 32;
  unsigned int hindex;
  size_t pat_s1 = 0, pat_s2 = 0;

  memset(buf1, 32, 2 * DPS_LM_MAXGRAM1 + 1);
  memset(buf2, 32, 2 * DPS_LM_MAXGRAM2 + 1);
  
  for (; text <= end; text++) {
    code = (unsigned char)(*text);
    if ((code <= 32) && (code >= 8) && (prevb <= 32) && (prevb >= 8)) continue;
    buf1[pat_s1] = buf1[pat_s1 + DPS_LM_MAXGRAM1] = buf2[pat_s2] = buf2[pat_s2 + DPS_LM_MAXGRAM2] = prevb = code;
    hindex = DpsHash32(buf1 + pat_s1 + 1, DPS_LM_MAXGRAM1) & DPS_LM_HASHMASK;
    map->memb3[hindex].count++;
    hindex = DpsHash32(buf2 + pat_s2 + 1, DPS_LM_MAXGRAM2) & DPS_LM_HASHMASK;
    map->memb6[hindex].count++;
    pat_s1++; pat_s2++;
    if (pat_s1 == DPS_LM_MAXGRAM1) pat_s1 = 0;
    if (pat_s2 == DPS_LM_MAXGRAM2) pat_s2 = 0;
    map->nbytes++;
    if (max_nbytes > 0 && map->nbytes >= max_nbytes) break;
  }
}



int DpsLMstatcmp(const void * i1, const void * i2) {
  register const DPS_MAPSTAT *s1 = (const DPS_MAPSTAT *)i1;
  register const DPS_MAPSTAT *s2 = (const DPS_MAPSTAT *)i2;

  if (s1->miss < s2->miss) return -1;
  if (s1->miss > s2->miss) return 1;
  if (s1->hits < s2->hits) return -1;
  if (s1->hits > s2->hits) return 1;

  return 0;
}

int DpsLMcmpCount(const void * i1, const void * i2) {
     register const DPS_LANGITEM *m1 = i1;
     register const DPS_LANGITEM *m2 = i2;
     if (m2->count > m1->count) return 1;
     if (m2->count < m1->count) return -1;
     return 0;
}

int DpsLMcmpIndex(const void * i1, const void * i2) {
     register const DPS_LANGITEM * m1 = i1;
     register const DPS_LANGITEM * m2 = i2;
     if (m2->index > m1->index) return -1;
     if (m2->index < m1->index) return 1;
     return 0;
}


void DpsPrepareLangMap(DPS_LANGMAP * map){
     DpsPreSort(map->memb3, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpCount);
     DpsPreSort(map->memb3, DPS_LM_TOPCNT, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
     DpsPreSort(map->memb6, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpCount);
     DpsPreSort(map->memb6, DPS_LM_TOPCNT, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
}


void DpsCheckLangMap(DPS_LANGMAP * map0, DPS_LANGMAP * map1, DPS_MAPSTAT *Stat, size_t InfMiss) {
     register int i;
     DPS_LANGITEM *HIT;

     Stat->hits = Stat->miss = 0;
     for (i = 0; i < DPS_LM_TOPCNT; i++) {
       if ( (HIT = bsearch(&map1->memb3[i], map0->memb3, DPS_LM_TOPCNT, sizeof(DPS_LANGITEM), &DpsLMcmpIndex)) == NULL) {
	 Stat->miss++;
       } else {
/*	 Stat->hits++;*/
	 register int p = (HIT - map0->memb3);
	 Stat->hits += (i >= p) ? (i - p) : (p - i);
       }
       if ( (HIT = bsearch(&map1->memb6[i], map0->memb6, DPS_LM_TOPCNT, sizeof(DPS_LANGITEM), &DpsLMcmpIndex)) == NULL) {
	 Stat->miss++;
       } else {
/*	 Stat->hits++;*/
	 register int p = (HIT - map0->memb6);
	 Stat->hits += (i >= p) ? (i - p) : (p - i);
       }
       if (Stat->miss > InfMiss) break;
     }
}


int  DpsGuessCharSet(DPS_AGENT *Indexer, DPS_DOCUMENT * Doc,DPS_LANGMAPLIST *List,DPS_LANGMAP *LangMap){
     DPS_CHARSET * cs;
     DPS_MAPSTAT * mapstat = NULL;
     DPS_LANGMAP *Cmap;
     const char *server_lang = DpsStrdup(DpsVarListFindStr(&Doc->Sections, "Content-Language", ""));
     const char *lang = "";
     const char *meta_lang = DpsStrdup(DpsVarListFindStr(&Doc->Sections, "Meta-Language", ""));
     const char *server_charset =  DpsCharsetCanonicalName(DpsVarListFindStr(&Doc->Sections, "Server-Charset", ""));
     const char *meta_charset =  DpsCharsetCanonicalName(DpsVarListFindStr(&Doc->Sections, "Meta-Charset", ""));
     const char *charset =  DpsCharsetCanonicalName(DpsVarListFindStr(&Doc->Sections, "RemoteCharset", ""));
     size_t InfMiss = 2 * DPS_LM_TOPCNT + 4;
     size_t i;
     int have_server_lang = (*server_lang != '\0');
     int use_meta, u;
#ifdef DEBUG_GUESSER
     int u1, u2, u3, u4;
#endif
     
     use_meta = Indexer->Flags.use_meta;
			
#ifdef DEBUG_GUESSER
     DpsLog(Indexer, DPS_LOG_EXTRA, "Guesser start0: meta_lang: %s, meta_charset: %s", 
	    DPS_NULL2EMPTY(meta_lang), DPS_NULL2EMPTY(meta_charset));
     DpsLog(Indexer, DPS_LOG_EXTRA, "Guesser start0: server_lang: %s, server_charset: %s", 
	    DPS_NULL2EMPTY(server_lang), DPS_NULL2EMPTY(server_charset));
#endif

     if (charset == NULL) {
       if (server_charset != NULL 
	   && (!use_meta || meta_charset == NULL || !strcasecmp(server_charset, meta_charset) )) charset = server_charset;
       else if (use_meta && server_charset == NULL && meta_charset != NULL) charset = meta_charset;
     }
     DpsVarListReplaceStr(&Doc->Sections, "Charset", DPS_NULL2EMPTY(charset));

     if (*lang == '\0') {
       if (use_meta && *server_lang != '\0' && !strcasecmp(server_lang, meta_lang)) lang = server_lang;
       else if (*server_lang != '\0' && !use_meta) lang = server_lang;
       else if (*meta_lang != '\0' && use_meta) DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = meta_lang);
     }

#ifdef DEBUG_GUESSER
     DpsLog(Indexer, DPS_LOG_EXTRA, "Guesser start1: lang: %s, charset: %s", DPS_NULL2EMPTY(lang), DPS_NULL2EMPTY(charset));
#endif

      if (charset != NULL && *lang != '\0') {
	if (FindLangMap(List, lang, charset, NULL, 0) == NULL) {
	  charset = NULL;
	}
      }    

#ifdef DEBUG_GUESSER
     DpsLog(Indexer, DPS_LOG_EXTRA, "Guesser start2: lang: %s, charset: %s", DPS_NULL2EMPTY(lang), DPS_NULL2EMPTY(charset));
#endif
     
     /* 
          TODO for Guesser
          
          There are three sources to detect charset:
          1. HTTP header:  Content-Type: ... charset=XXX
          2. <META HTTP-EQUIV="Content-Type" Contnet="... charset=YYY">
          3. ZZZ[i] - array of guessed charsets in mapstat[]
          good(ZZZ[n]) means that guesser returned good results for n.

          Now we may have various combinations:
          Simple situations, non-guessed and guessed charsets
          seem to be the same. At least one of non-guessed
          charset is the same with the best guessed charset
          and guessed charset gave good results:

          1. XXX==YYY==ZZZ[0],      good(ZZZ[0]). Take XXX value.
          2. No YYY, XXX=ZZZ[0] and good(ZZZ[0]). Take XXX value.
          3. No XXX, YYY=ZZZ[0] and good(ZZZ[0]). Take YYY value.
          4. XXX<>YYY, XXX==ZZZ[0], good(ZZZ[0]). Take XXX value.
          5. XXX<>YYY, YYY==ZZZ[0], good(ZZZ[0]). Take XXX value.
               4 and 5 seem to be webmaster mistake.

          There are the same fith situations when ZZZ[x] is still good 
          enough, but it is not the best one, i.e. x>0 
          Choose charset in the same way.
     */

     DPS_GETLOCK(Indexer, DPS_LOCK_CONF);

     if (charset == NULL || *lang == '\0') {
	  charset = NULL; lang = "";

       /* Prepare document langmap */
       DpsPrepareLangMap(LangMap);


       /* Allocate memory for comparison statistics */
       if ((mapstat=(DPS_MAPSTAT *)DpsMalloc((List->nmaps + 1) * sizeof(DPS_MAPSTAT))) == NULL) {
	 DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);
	 DpsLog(Indexer, DPS_LOG_ERROR, "Can't alloc momory for DpsGuessCharSet (%d bytes)", List->nmaps*sizeof(DPS_MAPSTAT));
	 DPS_FREE(server_lang); DPS_FREE(meta_lang);
	 return DPS_ERROR;
       }
     
       for(i=0;i<List->nmaps;i++){
          mapstat[i].map=&List->Map[i];
          if ((charset == NULL) && (*lang == '\0')) {
            DpsCheckLangMap(&List->Map[i], LangMap, &mapstat[i], InfMiss);
          } else if ((charset != NULL) && 
	     (!strcasecmp(mapstat[i].map->charset, charset) || !strcasecmp(mapstat[i].map->charset, DPS_NULL2EMPTY(meta_charset))) ) {
            DpsCheckLangMap(&List->Map[i], LangMap, &mapstat[i], InfMiss);
          } else if ((*lang) && !strncasecmp(mapstat[i].map->lang, lang, strlen(mapstat[i].map->lang) )) {
            DpsCheckLangMap(&List->Map[i], LangMap, &mapstat[i], InfMiss);
          } else {
            mapstat[i].hits = 0;
	    mapstat[i].miss = DPS_LM_TOPCNT + 1;
          }
          if (mapstat[i].miss < InfMiss) InfMiss = mapstat[i].miss;
       }

       /* Sort statistics in quality order */
       if (List->nmaps > 1) DpsSort(mapstat, List->nmaps, sizeof(DPS_MAPSTAT), &DpsLMstatcmp);

       /* Display results, best is shown first */
/*
#ifdef DEBUG_GUESSER
       for (i = 0; i < 5; i++)
	 DpsLog(Indexer, DPS_LOG_EXTRA, "Guesser: %dh:%dm %s-%s\n", 
		mapstat[i].hits, mapstat[i].miss, mapstat[i].map->lang, mapstat[i].map->charset);
#endif
*/
       if (server_charset != NULL || meta_charset != NULL)
	 for(i=0;i<List->nmaps;i++){

#ifdef DEBUG_GUESSER
	   DpsLog(Indexer, DPS_LOG_EXTRA, "Guesser @ charset: %dh:%dm %s-%s\n", 
		  mapstat[i].hits, mapstat[i].miss, mapstat[i].map->lang, mapstat[i].map->charset);
#endif

	   if (mapstat[i].map->lang && (*lang != '\0') && (!strncasecmp(mapstat[i].map->lang, lang, strlen(mapstat[i].map->lang) ))) {
	     if(mapstat[i].map->charset && !strcasecmp(mapstat[i].map->charset, DPS_NULL2EMPTY(server_charset))) {
	       DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
               DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[i].map->lang);
	     } else if(mapstat[i].map->charset && !strcasecmp(mapstat[i].map->charset, DPS_NULL2EMPTY(meta_charset))) {
	       DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
               DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[i].map->lang);
	     }
	   } else {
	     if(mapstat[i].map->charset && !strcasecmp(mapstat[i].map->charset, DPS_NULL2EMPTY(server_charset))) {
	       DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
               DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[i].map->lang);
	     } else if(mapstat[i].map->charset && !strcasecmp(mapstat[i].map->charset, DPS_NULL2EMPTY(meta_charset))) {
	       DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
               DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[i].map->lang);
	     }
	   }
	   if (charset != NULL) break;
	 }
          
       for(i=0;i<List->nmaps;i++){

	 if ((*lang != '\0') && (charset != NULL)) break;
#ifdef DEBUG_GUESSER
	 DpsLog(Indexer, DPS_LOG_EXTRA, "Guesser @ first: %dh:%dm %s.%s\n", 
		mapstat[i].hits, mapstat[i].miss, mapstat[i].map->lang, mapstat[i].map->charset);
#endif

	 if(mapstat[i].map->lang && *lang == '\0'/* && (*charset == '\0' || (!strcasecmp(mapstat[i].map->charset, charset)))*/ ){
               DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[i].map->lang);
	 }
	 if (mapstat[i].map->charset && charset == NULL && (!strcmp(lang, mapstat[i].map->lang)) ) {
               DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
	 }
          
/*          fprintf(stderr, "Guesser: %dh:%dm %s-%s\n",mapstat[i].hits, mapstat[i].miss,mapstat[i].map->lang,mapstat[i].map->charset);*/
          
       }
       if (List->nmaps > 0 && mapstat[0].map->charset && (charset == NULL) ) {
	     DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[0].map->charset);
       }
       if (List->nmaps > 0 && mapstat[0].map->lang && (*lang == '\0') ) {
             DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[0].map->lang);
       }
       DPS_FREE(mapstat);
     }
     
    u = Indexer->Flags.update_lm;
#ifdef DEBUG_GUESSER
    u1 = u;
#endif

    if (u) {
      if (have_server_lang) {
	if (use_meta && strcasecmp(server_lang, meta_lang)) u = 0;
      } else {
	if (use_meta && *meta_lang == '\0') u = 0;
      }
#ifdef DEBUG_GUESSER
      u2 = u;
#endif
    }
    if (u) {
      if (server_charset != NULL) {
	if (meta_charset != NULL && strcasecmp(server_charset, meta_charset)) u = 0;
      } else {
	if (!use_meta || meta_charset == NULL) u = 0;
      }
#ifdef DEBUG_GUESSER
      u3 = u;
#endif
    }

#ifdef DEBUG_GUESSER
    DpsLog(Indexer, DPS_LOG_INFO, "u:%d <- u1:%d u2:%d u3:%d srvlng:%s mtlng:%s  srvcs:%s mtcs:%s", 
	   u, u1, u2, u3, server_lang, meta_lang, (server_charset) ? server_charset : "<NULL>", (meta_charset)?meta_charset:"<NULL>");
#endif    

    if (u) {

      Cmap = FindLangMap(&Indexer->Conf->LangMaps, (have_server_lang) ? server_lang : meta_lang, 
			 (server_charset != NULL) ? server_charset : meta_charset, NULL, 1);
       if (Cmap != NULL) {
/*	 DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = Cmap->lang);*/

	 DpsSort(Cmap->memb3, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
	 DpsSort(LangMap->memb3, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
	 DpsSort(Cmap->memb6, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
	 DpsSort(LangMap->memb6, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
	 for (i = 0; i < DPS_LM_HASHMASK+1; i++) {
#ifdef DEBUG_GUESSER
	   if (Cmap->memb[i].count == 0) {
	     strcpy(Cmap->memb[i].str, LangMap->memb[i].str);
	   }
#endif
	   Cmap->memb3[i].count += LangMap->memb3[i].count;
	   Cmap->memb6[i].count += LangMap->memb6[i].count;
	 }
	 DpsPrepareLangMap(Cmap);
	 Cmap->needsave = 1;
	 DpsLog(Indexer, 
#ifdef DEBUG_GUESSER
		DPS_LOG_INFO,
#else
		DPS_LOG_EXTRA, 
#endif
		"Lang map: %s.%s updated", Cmap->lang, Cmap->charset);
       }
    }
    DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);
    cs = DpsGetCharSet(DpsVarListFindStr(&Doc->Sections, "Charset", "iso8859-1"));
    if (cs) Doc->charset_id = cs->id;
    DPS_FREE(server_lang); DPS_FREE(meta_lang);
    return DPS_OK;
}
